from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import math
import scipy.stats as stats


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def grank(data):
    if type(data) == np.ndarray or type(data) == list:
        alldata = data.copy()
        data = data.copy()
    else:
        alldata = data.values.copy()
        data = data.values.copy()
    alldata.sort()
    tmp_df = pd.DataFrame({'value': alldata})
    tmp_df['rank'] = tmp_df.index + 1
    value_to_rank = tmp_df.groupby('value').mean().reset_index()
    samp = pd.DataFrame({'value': data})
    samp = pd.merge(samp, value_to_rank, how='left')
    return samp['rank']


def ranksum_z_test(df=None, to_compute='', alternative=None, precision=4, alpha=0.05):
    """
    df can only have two columns and df.shape[0] > 10
    alternative has three options: 'two-sided', 'less', 'greater'
    """
    # sort all data points by values
    tmp_values = df.values.reshape(-1)
    tmp_values = tmp_values[~np.isnan(tmp_values)]
    tmp_values.sort()

    # assign ranks
    updated_df = pd.DataFrame({'value': tmp_values})
    updated_df['rank'] = updated_df.index + 1

    # average rank for identical value
    updated_df = updated_df.groupby('value').mean().reset_index()
    # display(updated_df)

    # Compute Sum of Ranks
    samp1 = pd.DataFrame({'value': df[to_compute].dropna().values})
    samp1 = pd.merge(samp1, updated_df)
    T = samp1['rank'].sum()

    # compute mean and standard deviation
    n1 = df.iloc[:, 0].dropna().shape[0]
    n2 = df.iloc[:, 1].dropna().shape[0]

    E_T = n1*(n1+n2+1)/2

    sigmaT = (n1*n2*(n1+n2+1)/12) ** 0.5
    z = (T-E_T)/sigmaT
    # compute p-value
    # right (greater)
    p_value = 1 - stats.norm.cdf(z)

    if alternative == 'greater':
        pass
    elif alternative == 'less':
        p_value = stats.norm.cdf(z)
    elif alternative == 'two-sided':
        # two-tail
        if p_value > 0.5:
            p_value = stats.norm.cdf(z)
        p_value *= 2
    flag = False
    if p_value < alpha:
        flag = True

    result = f'''======= z-test =======
T (sum of ranks) = {T}
(n1, n2) = ({n1}, {n2})
mu_t = {E_T}
sigma_t = {sigmaT}
z statistic value (observed) = {z:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 ({alternative}) → {flag}
'''
    print(result)
    result_dict = {'T': T, 'ET': E_T,
                   'sigmaT': sigmaT, 'z': z, 'p-value': p_value}
    return updated_df, result_dict


def sign_binom_test(diff=None, sign='+', alternative=None, precision=4, alpha=0.05):
    n = diff.size - np.sum(diff == 0)

    if sign == '+':
        sign_count = np.sum(diff > 0)
    else:
        sign_count = np.sum(diff < 0)

    if alternative == 'greater' or alternative == 'less':
        # 如果超過一半就要切換
        if sign_count > n / 2:
            p_value = 1 - stats.binom.cdf(sign_count - 1, n=n, p=0.5)
        else:
            p_value = stats.binom.cdf(sign_count, n=n, p=0.5)
    elif alternative == 'two-sided':
        p_value = stats.binom.cdf(sign_count, n=n, p=0.5)
        if p_value > 0.5:
            p_value = 1 - stats.binom.cdf(sign_count - 1, n=n, p=0.5)

        p_value *= 2

    flag = False
    if p_value < alpha:
        flag = True

    result = f'''======= Sign Test - Binomial Distribution =======
(For small sample size (<= 10))

Targeted Sign: {sign}
n = {n}
Sign counts = {sign_count}

p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 ({alternative}) → {flag}
    '''
    print(result)
    return sign_count, p_value


def sign_z_test(diff=None, sign='+', alternative=None, precision=4, alpha=0.05):
    diff = diff[~(diff == 0)]
    n = len(diff)

    if sign == '+':
        T = np.sum(diff > 0)
    else:
        T = np.sum(diff < 0)
    z_stat = (T - 0.5 * n) / (.5 * (n ** 0.5))
    # right tail
    if alternative == 'greater':
        p_value = 1 - stats.norm.cdf(z_stat)
    elif alternative == 'less':
        p_value = stats.norm.cdf(z_stat)
    elif alternative == 'two-sided':
        p_value = 1 - stats.norm.cdf(z_stat)
        if p_value > 0.5:
            p_value = stats.norm.cdf(z_stat)
        p_value *= 2
    flag = False
    if p_value < alpha:
        flag = True
    result = f'''======= Sign Test - z Statistic =======
(For large sample size (> 10))

Targeted Sign: {sign}
n = {n}
Sign counts = {T}

z statistic = {z_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 ({alternative}) → {flag}
    '''
    print(result)

    return T, p_value


def wilcoxon_signed_ranksum_z_test(diff=None, sign='+', alternative=None, precision=4, alpha=0.05):

    diff = diff[~(diff == 0)]
    n = len(diff)

    diff_abs = np.sort(np.abs(diff).to_numpy())

    updated_diff = pd.DataFrame({'diff_abs': diff_abs})
    updated_diff['rank'] = updated_diff.index + 1
    updated_diff = updated_diff.groupby('diff_abs').mean().reset_index()

    new_df = pd.DataFrame({'diff': diff, 'diff_abs': np.abs(diff)})
    new_df = pd.merge(new_df, updated_diff)

    if sign == '+':
        T = np.sum(new_df['rank'][new_df['diff'] > 0])
    else:
        T = np.sum(new_df['rank'][new_df['diff'] < 0])

    E_T = n * (n + 1) / 4
    sigma_T = (n * (n + 1) * (2 * n + 1) / 24) ** 0.5

    z_stat = (T - E_T) / sigma_T

    if alternative == 'greater':
        # right tail test
        p_value = 1 - stats.norm.cdf(z_stat)
    elif alternative == 'less':
        # left tail test
        p_value = stats.norm.cdf(z_stat)
    elif alternative == 'two-sided':
        # two-tailed test
        p_value = 1 - stats.norm.cdf(z_stat)
        if p_value > 0.5:
            p_value = stats.norm.cdf(z_stat)
        p_value *= 2

    flag = False
    if p_value < alpha:
        flag = True

    result = f'''======= Wilcoxon Signed Rank Sum Test - z Statistic =======
(For large sample size (> 30))

Targeted Sign: {sign}
n = {n}
Sum of rank (T statistic) = {T}

mu_t = {E_T}
sigma_t = {sigma_T}

z statistic value (observed) = {z_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 ({alternative}) → {flag}
    '''
    print(result)

    result_dict = {'n': n, 'T': T, 'E_T': E_T,
                   'sigma_T': sigma_T, 'z_stat': z_stat, 'p_value': p_value}

    return new_df, result_dict


def kruskal_chi2_test(data=None, alpha=0.05, precision=4):
    """
    col = 要比較的 target
    row = data for each target
    """
    if type(data) == pd.DataFrame:
        data = data.copy().to_numpy()
        alldata = np.concatenate(data.copy())
    else:
        alldata = np.concatenate(data.copy())

    k = data.shape[1]
    alldata.sort()

    tmp_df = pd.DataFrame(({'value': alldata}))
    tmp_df['rank'] = tmp_df.index + 1  # rank
    value_to_rank = tmp_df.groupby('value').mean().reset_index()
    T = []
    sample_rank_df = []
    for i in range(k):

        samp = pd.DataFrame(
            {'value': data[:, i][~np.isnan(data[:, i])]})

        samp = pd.merge(samp, value_to_rank)
        sample_rank_df.append(samp)
        T.append(samp['rank'].sum())

    n = [len(data[:, i][~np.isnan(data[:, i])]) for i in range(k)]

    # print(T)
    # print(n)

    rule_of_five_str = ""
    if (np.sum(np.array(n) < 5) > 0):
        rule_of_five_str += "!(At least one sample size is less than 5)"
    else:
        rule_of_five_str += "(All sample size >= 5)"

    N = np.sum(n)

    t_over_n = 0

    for i in range(k):
        t_over_n += T[i] ** 2 / n[i]

    H = 12 / N / (N + 1) * t_over_n - 3 * (N + 1)
    p_value = 1 - stats.chi2.cdf(H, k - 1)
    chi2_stat = stats.chi2.ppf(1 - alpha, k - 1)

    result_dict = {'H': H, 'p-value': p_value,
                   'T': T, 'sample_rank_df': sample_rank_df}
    flag = p_value < alpha

    result = f'''======= Kruskal-Wallis Test with Chi-squared Test =======
{rule_of_five_str}

H statistic value (observed) = {H:.{precision}f}
chi2 critical value = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 (Not all {k} population locations are the same) → {flag}
    '''
    print(result)
    return result_dict


def friedman_chi2_test(data=None, alpha=0.05, precision=4):
    """
    col = 要比較的 target
    row = blocked data for each target
    """
    if type(data) == np.ndarray:
        data = pd.DataFrame(data)

    new_df = data.apply(grank, axis=1)
    b, k = new_df.shape

    rule_of_five_str = ""
    if (b < 5 and k < 5):
        rule_of_five_str += f"!(Number of blocks = {b} < 5 and number of populations = {k} < 5)"
    else:
        rule_of_five_str += f"(Number of blocks = {b} >= 5 or number of populations {k} >= 5)"

    T = new_df.sum().to_numpy()

    F_r = 12 / b / k / (k + 1) * np.sum(T ** 2) - 3 * b * (k + 1)
    p_value = 1 - stats.chi2.cdf(F_r, k - 1)
    chi2_stat = stats.chi2.ppf(1 - alpha, k - 1)

    result_dict = {'F_r': F_r, 'p-value': p_value,
                   'T': T, 'sample_ranked_df': new_df}
    flag = p_value < alpha

    result = f'''======= Friedman Test with Chi-squared Test =======
{rule_of_five_str}

F_r statistic value (observed) = {F_r:.{precision}f}
chi2 critical value = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 (Not all {k} population locations are the same) → {flag}
    '''
    print(result)
    return result_dict


def pearson_test(data=None, a=None, b=None, alpha=0.05, precision=4):
    """
    a, b 還不能傳入東西
    Make sure that data is in the form of [a, b]
    """
    cov_mat = np.cov(data.values, rowvar=False)
    cor_mat = np.corrcoef(data.values, rowvar=False)
    cov = cov_mat[0][1]
    cor = cor_mat[0][1]

    n = data.shape[0]
    d_of_f = n - 2
    t_c = stats.t.ppf(1 - alpha / 2, df=d_of_f)
    t_stat = cor * (((n - 2) / (1 - cor ** 2)) ** 0.5)

    flag = abs(t_stat) > t_c
    result_dict = {'cov': cov, 't_stat': t_stat, 'cor': cor, 't_c': t_c}
    results = f"""======= Pearson Correlation Coefficient =======
Covariance: {cov:.{precision}f}
Coefficient of Correlation: {cor:.{precision}f}

t (Critical Value) = {t_c:.{precision}f}
t (Observed Value) = {t_stat:.{precision}f}

Reject H_0 (There are linear relationship between two variables) → {flag}
"""

    print(results)

    return result_dict


def spearman_test(a=None, b=None, alpha=0.05, precision=4):
    spearman_restult_cor, spearman_restult_p_value = stats.spearmanr(a, b)
    # print(f'Correlation = {cor:.4f}, p-value={p_value:.4f}')
    n = len(a)

    rule_of_30_str = ''

    results = f"""======= Spearman Rank Correlation Coefficient =======
[scipy.stats.spearmanr]
Coefficient of Correlation: {spearman_restult_cor:.{precision}f}
p-value={spearman_restult_p_value:.{precision}f} ({inter_p_value(spearman_restult_p_value)})
"""

    if (n < 30):
        rule_of_30_str += f"!(n = {n} < 30)"
        flag = spearman_restult_p_value < alpha
        results += f"""
Reject H_0 (There are relationship between two variables) → {flag}
        """
        result_dict = {'spearman_result': [
            spearman_restult_cor, spearman_restult_p_value]}
    else:
        rule_of_30_str += f"(n = {n} >= 30)"
        flag = spearman_restult_p_value < alpha
        results += f"""
Reject H_0 (There are relationship between two variables) → {flag}
        """
        z_stat = spearman_restult_cor * ((n - 1) ** 0.5)
        z_cv = stats.norm.ppf(1 - alpha/2)
        p_value = stats.norm.sf(z_stat) * 2
        if p_value > 1:
            p_value = stats.norm.cdf(z_stat) * 2
        flag = p_value < alpha
        results += f"""
[z test statistic]
{rule_of_30_str}

r_s: {spearman_restult_cor:.{precision}f} (using spearmanr's result)
z stat (observed value) = {z_stat:.{precision}f}
z (critical value) = {z_cv:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 (There are relationship between two variables) → {flag}
        """

        result_dict = {'spearman_result': [
            spearman_restult_cor, spearman_restult_p_value], 'z_stat': z_stat, 'z_cv': z_cv, 'p-value': p_value}

    print(results)

    return result_dict



from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns
import numpy as np
import math
from io import StringIO
import scipy.stats as stats
from scipy.stats import norm
import statsmodels.api as sm
import statsmodels.stats.api as sms
import statsmodels.formula.api as smf

_raw_r_table = StringIO(""",5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20
2,,,,,,,,"(2, 6)","(2, 6)","(2, 6)","(2, 6)","(2, 6)","(2, 6)","(2, 6)","(2, 6)","(2, 6)"
3,,"(2, 8)","(2, 8)","(2, 8)","(2, 8)","(2, 8)","(2, 8)","(2, 8)","(2, 8)","(2, 8)","(3, 8)","(3, 8)","(3, 8)","(3, 8)","(3, 8)","(3, 8)"
4,"(2, 9)","(2, 9)","(2, 10)","(3, 10)","(3, 10)","(3, 10)","(3, 10)","(3, 10)","(3, 10)","(3, 10)","(3, 10)","(4, 10)","(4, 10)","(4, 10)","(4, 10)","(4, 10)"
5,"(2, 10)","(3, 10)","(3, 11)","(3, 11)","(3, 12)","(3, 12)","(4, 12)","(4, 12)","(4, 12)","(4, 12)","(4, 12)","(4, 12)","(4, 12)","(5, 12)","(5, 12)","(5, 12)"
6,,"(3, 11)","(3, 12)","(3, 12)","(4, 13)","(4, 13)","(4, 13)","(4, 13)","(5, 14)","(5, 14)","(5, 14)","(5, 14)","(5, 14)","(6, 14)","(6, 14)","(6, 14)"
7,,,"(3, 13)","(4, 13)","(4, 14)","(5, 14)","(5, 14)","(5, 14)","(5, 15)","(5, 15)","(6, 15)","(6, 16)","(6, 16)","(6, 16)","(6, 16)","(6, 16)"
8,,,,"(4, 14)","(5, 14)","(5, 15)","(5, 15)","(6, 16)","(6, 16)","(6, 16)","(6, 16)","(6, 17)","(7, 17)","(7, 17)","(7, 17)","(7, 17)"
9,,,,,"(5, 15)","(5, 16)","(6, 16)","(6, 16)","(6, 17)","(7, 17)","(7, 18)","(7, 18)","(7, 18)","(8, 18)","(8, 18)","(8, 18)"
10,,,,,,"(6, 16)","(6, 17)","(7, 17)","(7, 18)","(7, 18)","(7, 18)","(8, 19)","(8, 19)","(8, 19)","(8, 20)","(9, 20)"
11,,,,,,,"(7, 17)","(7, 18)","(7, 19)","(8, 19)","(8, 19)","(8, 20)","(9, 20)","(9, 20)","(9, 21)","(9, 21)"
12,,,,,,,,"(7, 19)","(8, 19)","(8, 20)","(8, 20)","(9, 21)","(9, 21)","(9, 21)","(10, 22)","(10, 22)"
13,,,,,,,,,"(8, 20)","(9, 20)","(9, 21)","(9, 21)","(10, 22)","(10, 22)","(10, 23)","(10, 23)"
14,,,,,,,,,,"(9, 21)","(9, 22)","(10, 22)","(10, 23)","(10, 23)","(11, 23)","(11, 24)"
15,,,,,,,,,,,"(10, 22)","(10, 23)","(11, 23)","(11, 24)","(11, 24)","(12, 25)"
16,,,,,,,,,,,,"(11, 23)","(11, 24)","(11, 25)","(12, 25)","(12, 25)"
17,,,,,,,,,,,,,"(11, 25)","(12, 25)","(12, 26)","(13, 26)"
18,,,,,,,,,,,,,,"(12, 26)","(13, 26)","(13, 27)"
19,,,,,,,,,,,,,,,"(13, 27)","(13, 27)"
20,,,,,,,,,,,,,,,,"(14, 28)"
""")
_runs_test_table = pd.read_csv(_raw_r_table, index_col=0)


def add_margin(ax, x=0.05, y=0.05):

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    xmargin = (xlim[1]-xlim[0])*x
    ymargin = (ylim[1]-ylim[0])*y

    ax.set_xlim(xlim[0]-xmargin, xlim[1]+xmargin)
    ax.set_ylim(ylim[0]-ymargin, ylim[1]+ymargin)


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def c_of_c_test(r, n, beta1=0, alpha=.05, precision=4, show=True, option='two-tail', **kwargs):
    """
    two-tail, right, left
    kwargs needs to contain b1 and std_err if used
    """
    opt = option.lower()[0]
    t_value = r * ((n-2)/(1 - r**2)) ** 0.5
    if beta1 != 0:
        t_value = (kwargs['b1'] - beta1) / kwargs['std_err']
    t_critical = stats.t.ppf(1 - alpha/2, n - 2)
    p_value = stats.t.sf(np.abs(t_value), n - 2)*2
    if opt == 't':
        # two-tail test
        option = 'Two-Tail Test'
        p_value = (1 - stats.t.cdf(t_value, df=n-2)) * 2
        if (p_value > 1):
            p_value = (stats.t.cdf(t_value, df=n-2)) * 2
        t_critical = stats.t.ppf(1 - alpha/2, df=n-2)
    else:
        if opt == 'l':
            option = 'One-Tail Test (left tail)'
            p_value = stats.t.cdf(t_value, df=n - 2)
            t_critical = stats.t.ppf(alpha, df=n - 2)
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_value = stats.t.sf(t_value, df=n - 2)
            t_critical = stats.t.ppf(1 - alpha, df=n - 2)

    des = f"""r = {r:.{precision}f}
t (critical value) = {t_critical:.{precision}f}
t (observed value)  = {t_value:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
    """
    if show:
        print(des)
    return t_value, t_critical, p_value, option


def SimpleLinearRegression(Independence=None, Dependence=None, df=None, alpha=0.05, precision=4, show_summary=True, plot=False, slope_option='two-tail', beta1=0, coeff_option='two-tail', kwargs={'x': 0.02, 'y': 0.00, 'title': 'Scatter Plot'}):
    """
    df_result = smf.ols(f'{Dependence}~ {Independence}', data=df).fit()
    """
    slope, intercept, r_value, p_value, std_err_sb1 = stats.linregress(
        df[Independence], df[Dependence])
    flag = p_value < alpha
    result_dict = {'slope': slope, 'intercept': intercept,
                   'r_value': r_value, 'p_value': p_value}
    fig, ax = plt.subplots()

    ax = sns.regplot(x=Independence, y=Dependence, data=df, ci=None, scatter_kws={'color': 'dodgerblue'}, line_kws={
        'color': '#ffaa77', 'label': f"y = {intercept:.4f} + {slope:.4f} x"})

    add_margin(ax, x=kwargs['x'], y=kwargs['y'])  # Call this after tsplot
    result_dict['fig'] = fig
    result_dict['ax'] = ax
    if plot:
        plt.legend()
        plt.title(kwargs['title'])
        if 'xlabel' in kwargs:
            plt.xlabel(kwargs['xlabel'])
        else:
            plt.xlabel(Independence)
        if 'ylabel' in kwargs:
            plt.ylabel(kwargs['ylabel'])
        else:
            plt.ylabel(Dependence)
        plt.show()
    else:
        plt.close(fig)

    df_result = smf.ols(f'{Dependence}~ {Independence}', data=df).fit()
    result_dict['result'] = df_result
    s_e = df_result.mse_resid ** 0.5
    ssr_value = df_result.ssr
    r_square = r_value ** 2
    s_t_value, s_t_critical, s_p_value, s_option = c_of_c_test(
        r_value, df_result.nobs, beta1=beta1, alpha=alpha, precision=precision, show=False, option=slope_option, kwargs={'b1': slope, 'std_err': std_err_sb1})
    s_flag = s_p_value < alpha
    t_t_value, t_t_critical, t_p_value, t_option = c_of_c_test(
        r_value, df_result.nobs, alpha=alpha, precision=precision, show=False, option=coeff_option)
    if t_option == 'Two-Tail Test':
        t_p_value = p_value

    ci_b1 = df_result.conf_int(alpha)[1:].values[0]
    result_dict['ci_b1'] = ci_b1

    if show_summary:
        print(df_result.summary())
        print()
    results = f"""======= Simple Linear Regression Results =======
Dep. Variable: {Dependence}
No. of Observations: {int(df_result.nobs)}
Standard Error: {s_e:.{precision}f}
SSR: {ssr_value:.{precision}f}
R-square: {r_square:.{precision}f}

Estimated model: y = {intercept:.{precision}f} + {slope:.{precision}f} x

**** t-Test of Slope <{s_option}> ****
t (observed value): {s_t_value:.{precision}f}
t (critical value): {s_t_critical:.{precision}f}
b1 (slope): {slope:.{precision}f}
p-value: {s_p_value:.{precision}f} ({inter_p_value(s_p_value)})

Reject H_0 (Has some kind of relationship between two variables) → {s_flag}

{(1-alpha) * 100}% confidence interval = [{ci_b1[0]:.4f}, {ci_b1[1]:.4f}]

**** t-Test of Correlation Coefficient <{t_option}> ****
t (observed value): {t_t_value:.{precision}f}
t (critical value): {t_t_critical:.{precision}f}
r: {r_value:.{precision}f}
p-value: {t_p_value:.{precision}f} ({inter_p_value(t_p_value)})
Reject H_0 (Has Linear Relationship) → {flag}"""
    print(results)
    result_dict['description'] = results

    return result_dict



from matplotlib import pyplot as plt
import pandas as pd
import numpy as np
import math
import scipy.stats as stats


def con_level(alpha, n, s, sigma, show=True, ignore=True):
    """
Input: alpha, n (sample size), s (sample standard deviation), sigma (population standard deviation, can ignore if ignore=True - default), show=True
Return the confidence level at alpha. Return a dictionary: {"lcl": lcl, "ucl": ucl}

+ `show`: default is `True`. Set to `False` to disable rendering.
    """
    df_v = n - 1
    con_coef = 1 - alpha

    chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
    chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)

    lcl = (df_v) * (s ** 2) / chi2_cv_u
    ucl = (df_v) * (s ** 2) / chi2_cv_l

    result = f"""{con_coef * 100:.1f}% Confidence Interval: [{lcl:.4f}, {ucl:.4f}]

S^2 (Sample Variance): {s ** 2:.4f}
Sample Size: {n}
"""
    if ignore == False:
        flag = sigma ** 2 >= lcl and sigma ** 2 <= ucl
        if flag:
            cover = "covers the real/population variance."
        else:
            cover = "does not cover the real/population variance."

        result += f'''sigma^2 (Population Variance) = {sigma ** 2:.4f}
The interval {cover}
'''
    if show:
        print(result)
    return {"lcl": lcl, "ucl": ucl}


def rejection_region_method(alpha, n, s, sigma, option='left', precision=4, show=True, ignore=False):
    """
    Input: alpha, n, s, sigma, option='left', precision=4, show=True, ignore=False
    Output: 
        if opt == 't':
            return s2_l, s2_u
        else:
            return s2_c
    """
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    if opt == 't':
        option = 'Two-Tail Test'
        chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
        chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)
        s2_u = chi2_cv_u * (sigma ** 2) / df_v
        s2_l = chi2_cv_l * (sigma ** 2) / df_v
        flag = var < s2_l or var > s2_u
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2_l (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
chi2_u (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}

Using {option}:
S^2 (var) =  {var:.{precision}f}
S_L^2 (Lower bound for the critical value) = {s2_l:.{precision}f}
S_U^2 (Upper bound for the critical value) = {s2_u:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2_l (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
chi2_u (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}

Using {option}:
S_L^2 (Lower bound for the critical value) = {s2_l:.{precision}f}
S_U^2 (Upper bound for the critical value) = {s2_u:.{precision}f}
            '''

    else:
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            chi2_value = df_v * (s ** 2) / (sigma ** 2)
            chi2_cv = stats.chi2.ppf(alpha, df_v)
            s2_c = chi2_cv * (sigma ** 2) / df_v
            flag = var < s2_c
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            chi2_value = df_v * (s ** 2) / (sigma ** 2)
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
            s2_c = chi2_cv * (sigma ** 2) / df_v
            flag = var > s2_c
        if not ignore:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2 (Critical value) = {chi2_cv:.{precision}f}

Using {option}:
S^2 (var) =  {var:.{precision}f}
S_C^2 (Critical value) = {s2_c:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= The Rejection Region Method =======
Significant Level (alpha) = {alpha:.{precision}f}
chi2 (Critical value) = {chi2_cv:.{precision}f}

Using {option}:
S_C^2 (Critical value) = {s2_c:.{precision}f}
            '''

    if show:
        print(result)

    if opt == 't':
        return s2_l, s2_u
    else:
        return s2_c


def testing_statistic_method(alpha, n, s, sigma, option='left', precision=4, ignore=False):
    """
    Input: alpha, n, s, sigma, option='left', precision=4, show=True, ignore=False
    Output: 
        if opt == 't':
            return chi2_value, chi2_cv_l, chi2_cv_u
        else:
            return chi2_value, chi2_cv
    """
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    chi2_value = df_v * (s ** 2) / (sigma ** 2)

    if opt == 't':
        option = 'Two-Tail Test'
        chi2_cv_u = stats.chi2.ppf(1 - alpha / 2, df=df_v)
        chi2_cv_l = stats.chi2.ppf(alpha / 2, df=df_v)
        flag = chi2_value < chi2_cv_l or chi2_value > chi2_cv_u

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2 (Observed value) =  {chi2_value:.{precision}f}
Chi2_L (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
Chi2_U (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}
Reject H_0 → {flag}
            '''
        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2_L (Lower bound for Chi2 critical value) = {chi2_cv_l:.{precision}f}
Chi2_U (Upper bound for Chi2 critical value) = {chi2_cv_u:.{precision}f}
            '''

    else:
        if opt == 'l':
            # left tail
            option = 'One-Tail Test (left tail)'
            chi2_cv = stats.chi2.ppf(alpha, df_v)
            flag = chi2_value < chi2_cv
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
            flag = chi2_value > chi2_cv

        if not ignore:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2 (Observed value) =  {chi2_value:.{precision}f}
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
Reject H_0 → {flag}
            '''

        else:
            result = f'''======= Testing Statistic Method =======
Significant Level (alpha) = {alpha:.{precision}f}

Using {option}:
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
            '''

    print(result)
    if opt == 't':
        return chi2_value, chi2_cv_l, chi2_cv_u
    else:
        return chi2_value, chi2_cv


def inter_p_value(p_value):
    # interpretation
    if p_value >= 0 and p_value < 0.01:
        inter_p = 'Overwhelming Evidence'
    elif p_value >= 0.01 and p_value < 0.05:
        inter_p = 'Strong Evidence'
    elif p_value >= 0.05 and p_value < 0.1:
        inter_p = 'Weak Evidence'
    elif p_value >= .1:
        inter_p = 'No Evidence'
    return inter_p


def p_value_method(alpha, n, s, sigma, option='left', precision=4):
    """
    Input: alpha, n, s, sigma, option='left', precision=4
    Output: chi2_stat, p_value
    """
    opt = option.lower()[0]
    df_v = n - 1
    var = s ** 2
    chi2_stat = df_v * (s ** 2) / (sigma ** 2)

    if opt == 't':
        # two-tail test
        option = 'Two-Tail Test'
        alphas = np.arange(0, 1, 0.001)
        chi2s = stats.chi2.ppf(alphas, df_v)
        med = np.median(chi2s)

        if chi2_stat > med:
            p_value = 2 * stats.chi2.sf(chi2_stat, df_v)
        else:
            p_value = 2 * stats.chi2.cdf(chi2_stat, df_v)

        flag = p_value < alpha
        sub_result = f'''Using {option}:
Chi2 Median under d.f. {df_v}: {med:.{precision}f}
Chi2 Stat (Observed value) = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''
    else:
        if opt == 'l':
            option = 'One-Tail Test (left tail)'
            p_value = stats.chi2.cdf(chi2_stat, df_v)
            chi2_cv = stats.chi2.ppf(alpha, df_v)
        elif opt == 'r':
            option = 'One-Tail Test (right tail)'
            p_value = stats.chi2.sf(chi2_stat, df_v)
            chi2_cv = stats.chi2.ppf(1 - alpha, df_v)
        flag = p_value < alpha
        sub_result = f'''Using {option}:
Chi2_C (Critical value) = {chi2_cv:.{precision}f}
Chi2 Stat (Observed value) = {chi2_stat:.{precision}f}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})
Reject H_0 → {flag}
        '''

    result = f"""======= p-value Method =======
S^2 (var) = {var:.{precision}f}
Number of Observation = {n:.{precision}f}
Hypothesized Std. (H0 Std.) = {sigma:.{precision}f}
Sample Standard Deviation = {s:.{precision}f}
Significant Level (alpha) = {alpha:.{precision}f}

""" + sub_result

    print(result)

    return chi2_stat, p_value

# Chi-Squared Tests for two or more variables


def multinomial(freq_o, freq_e, alpha=0.05, precision=4):
    """
    >>> return chi2_stat, chi2_cv, p_value
    Check RULE of FIVE before running the test
    >>> if np.sum(freq_e < 5) > 0:
    >>>    print("Rule of five is not met. ")
    """
    # if np.sum(freq_e < 5) > 0:
    #     print("Rule of five is not met.")
    #     return
    stat, p_value = stats.chisquare(freq_o, freq_e)
    df = freq_o.shape[0]-1
    chi2_cv = stats.chi2.ppf(1 - alpha, df)
    flag = False
    if p_value < 0.05:
        flag = True
    result = f"""======= Goodness of Fit Test: A Multinomial Population =======
Chi-squared test: 
chi2 statistics = {stat:.{precision}f}
chi2 critical value = {chi2_cv:.{precision}f}
Degree of freedom = {df}
p-value = {p_value:.{precision}f} ({inter_p_value(p_value)})

Reject H_0 (does not follow the probability) → {flag}
"""
    print(result)
    return stat, chi2_cv, p_value


def contingency(cont_table=None, alpha=0.05, precision=4):
    """
    Check RULE of FIVE before running the test
    >>> return chi2, p, dof, ex
    """
    chi2, p, dof, ex = stats.chi2_contingency(cont_table, correction=False)
    chi2_cv = stats.chi2.ppf(1 - alpha, dof)
    flag = False
    if p < alpha:
        flag = True

    result = f"""======= Tests of Independence: Contingency Table =======
chi2 Statistics = {chi2:.{precision}f}
chi2 critical value = {chi2_cv:.{precision}f}

Degree of freedom = {dof}
p-value = {p:.{precision}f} ({inter_p_value(p)})

Reject H_0 (dependent) → {flag}

Expected Frequency:
{ex}
"""
    print(result)
    return chi2, p, dof, ex

# Poisson


def frequency_table(interval, observed, nobs, total_time, mu):
    o_vs_e_freq_df = pd.DataFrame(columns=['i', 'f_i', 'e_i', 'f_i - e_i'])
    poisson_table = stats.poisson(mu)
    N = total_time
    for (i, f) in zip(interval, observed):
        e = poisson_table.pmf(i)
        e *= N
        o_vs_e_freq_df = o_vs_e_freq_df.append(
            {'i': i, 'f_i': f, 'e_i': e, 'f_i - e_i': f - e}, ignore_index=True)
    return o_vs_e_freq_df


def df_combine(df=None, to_combine=None):
    """
    >>> to_combine = [[0,1,2],[10, 11, 12]]
    >>> updated_df = df_combine(df=o_vs_e_freq_df, to_combine=to_combine)
    """
    if len(to_combine) == 0:
        return df
    else:
        updated_df = pd.DataFrame(columns=['i', 'f_i', 'e_i', 'f_i - e_i'])
        n = len(to_combine)
        for i in range(n):
            if i == 0:
                current_end = to_combine[i][-1]
                if i + 1 < n:
                    next_start = to_combine[i + 1][0]
                else:
                    next_start = None
                updated_df = updated_df.append(
                    df.iloc[to_combine[i]].sum(axis=0), ignore_index=True)
                updated_df.i[i] = f'{to_combine[i]}'
                updated_df = updated_df.append(
                    df.iloc[current_end + 1:next_start], ignore_index=True)
            else:
                current_start = to_combine[i][0]
                prev_end = to_combine[i - 1][-1]
                current_end = to_combine[i][-1]
                if i + 1 < n:
                    next_start = to_combine[i + 1][0]
                else:
                    next_start = None
                updated_df = updated_df.append(
                    df.iloc[to_combine[i]].sum(axis=0), ignore_index=True)
                updated_df.iloc[current_start - prev_end +
                                (i - 1), 0] = f'{to_combine[i]}'
                updated_df = updated_df.append(
                    df.iloc[current_end + 1:next_start], ignore_index=True)

        return updated_df


def poisson_test(df, alpha=0.05, precision=4):
    """
    >>> return chi2_stat, chi2_cv, dof
    """
    dof = df.shape[0] - 1 - 1
    chi2_stat = sum((df['f_i - e_i']) ** 2 / df['e_i'])
    chi2_cv = stats.chi2.ppf(1 - alpha, df=dof)
    flag = False  # cannot reject
    if chi2_stat > chi2_cv:
        flag = True

    result = f"""======= Goodness of Fit Test: Poisson Test =======
chi2 statistic (observed value): {chi2_stat:.{precision}f}
chi2 critical value: {chi2_cv:.{precision}f}
Degree of freedom: {dof}

Reject H_0 (does not follow a Poisson distribution) → {flag}
"""
    print(result)
    return chi2_stat, chi2_cv, dof


def normal_test(srs=None, z_intervals=None, freq_o=None, alpha=0.05, precision=4):
    """
    pass in Series instead of DataFrame
    >>> return o_vs_e_freq_df, chi2_stat, chi2_cv, dof

    Note:
    >>> z_value = stats.norm.ppf(area)
    """
    if z_intervals is not None:
        o_vs_e_freq_df = pd.DataFrame(columns=['i', 'f_i', 'e_i', 'f_i - e_i'])

        # intervals = [(-1), (-1, 0), (0, 1), (1)]
        # freq_o = np.array([6, 27, 14, 3])
        nobs = np.sum(freq_o)

        snd = stats.norm

        for i, interval in enumerate(z_intervals):
            try:
                if len(interval) > 1:
                    z_value = snd.cdf(interval[1]) - snd.cdf(interval[0])
            except:
                if i == 0:
                    z_value = snd.cdf(interval)
                else:
                    z_value = snd.sf(interval)

            e_i = nobs * z_value
            o_vs_e_freq_df = o_vs_e_freq_df.append(
                {'i': interval, 'f_i': freq_o[i], 'e_i': e_i, 'f_i - e_i': freq_o[i] - e_i}, ignore_index=True)

        if np.sum(o_vs_e_freq_df['e_i'] < 5) > 0:
            print("Rule of five is not met. ")

        chi2_stat = np.sum(o_vs_e_freq_df['f_i - e_i']
                           ** 2 / o_vs_e_freq_df['e_i'])
        dof = o_vs_e_freq_df.shape[0] - 2 - 1
        chi2_cv = stats.chi2.ppf(1 - alpha, df=dof)
        flag = False
        if chi2_stat > chi2_cv:
            flag = True
        result = f"""======= Goodness of Fit Test: Normal Distribution =======
chi2 statistics = {chi2_stat:.{precision}f}
chi2 critical value = {chi2_cv:.{precision}f}
Degree of freedom = {dof}

Reject H_0 (reject the assumption that the population is normally distributed) → {flag}
    """
        print(result)

        return o_vs_e_freq_df, chi2_stat, chi2_cv, dof

    elif srs is not None:
        snd = stats.norm
        arr = np.array(srs)
        o_vs_e_freq_df = pd.DataFrame(columns=['i', 'f_i', 'e_i', 'f_i - e_i'])
        n = len(arr)
        mu = arr.mean()
        std = arr.std(ddof=1)
        interval = n / 5
        area = 1 / interval
        acc_area = area
        ctr = 0
        while abs(acc_area - 1) > 0.00000001:
            z_value = snd.ppf(acc_area)
            x_value = z_value * std + mu
            acc_area += area
            f_i = sum(arr <= x_value)
            arr = arr[arr > x_value]

            ctr += 1
            o_vs_e_freq_df = o_vs_e_freq_df.append(
                {'i': x_value, 'f_i': f_i, 'e_i': 5, 'f_i - e_i': f_i - 5}, ignore_index=True)

        num_df = o_vs_e_freq_df.i
        f_i = sum(arr > x_value)
        x_value = f'> {x_value:.3f}'
        o_vs_e_freq_df = o_vs_e_freq_df.append(
            {'i': x_value, 'f_i': f_i, 'e_i': 5, 'f_i - e_i': f_i - 5}, ignore_index=True)

        for i in range(1, ctr):
            o_vs_e_freq_df.iat[i, 0] = f'{num_df[i - 1]:.3f} ~ {num_df[i]:.3f}'
        o_vs_e_freq_df.iat[0, 0] = f'<= {o_vs_e_freq_df.i[0]:.3f}'

        # test
        chi2_stat = np.sum(o_vs_e_freq_df['f_i - e_i']
                           ** 2 / o_vs_e_freq_df['e_i'])
        dof = o_vs_e_freq_df.shape[0] - 2 - 1
        chi2_cv = stats.chi2.ppf(1 - alpha, df=dof)
        flag = False
        if chi2_stat > chi2_cv:
            flag = True
        result = f"""======= Goodness of Fit Test: Normal Distribution =======
chi2 statistics = {chi2_stat:.{precision}f}
chi2 critical value = {chi2_cv:.{precision}f}
Degree of freedom = {dof}

μ = {mu:.{precision}f}
σ = {std:.{precision}f}

Reject H_0 (reject the assumption that the population is normally distributed with μ = {mu:.{precision}f} and σ = {std:.{precision}f}) → {flag}
        """
        print(result)

        return o_vs_e_freq_df, chi2_stat, chi2_cv, dof
